import os
import re
import sys
import random


readme = """
===================================================================================
add detailed readme here to be printed out if YML file does not exist in the folder
...................................................................................
===================================================================================
"""

ymlTemplate = """

system: ... # options: mac, lin, win;
reviewer: eis100researcher # change this to your name;
path_to_kate: kate # add absolute path to Kate on your machine; or a working alias
files_to_open: 10 # the script will open 10 files by default; you can increase/decrease this value;
miu_main_path: ./data/ # the script should be run from inside the repository with MIU files
miu_specific_uri: ... # you can paste a URI of a specific book in order to focus on it

"""


def readCustomYML(text):
    text = text.strip()
    text = text.split("\n")

    dic = {}

    for t in text:
        t = t.split(":")
        key = t[0]
        val = re.sub("#.*$", "", t[1]).strip()

        dic[key] = val

    return dic


def generateDicOfMIUs(path):
    dic = {}
    for subdir, dirs, files in os.walk(path):
        for file in files:
            # process publication tf data
            if file.endswith(".EIS1600"):
                key = file
                value = os.path.join(subdir, file)
                dic[key] = value
    return(dic)


def sampleMIUs(dictionary, filterParameter, filesToOpen):
    selection = []

    for k in dictionary.keys():
        if filterParameter in k:
            selection.append(k)

    finalSelection = random.sample(selection, filesToOpen*3)

    return(finalSelection)


def main():
    if os.path.exists("file_picker.yml"):
        with open("file_picker.yml", "r", encoding="utf8") as f1:
            ymlText = f1.read().strip()
            ymlDic = readCustomYML(ymlText)

            print("===================================================================================")
            for k, v in ymlDic.items():
                print(k, ": ", v)
            print("===================================================================================")
    else:
        with open("file_picker.yml", "w", encoding="utf8") as f1:
            f1.write(ymlTemplate)

        print("===================================================================================")
        print("YML FILE HAS BEEN GENERATED; PLEASE UPDATE THE YML FILE AS DESCRIBED ==============")
        print("===================================================================================")
        print(ymlTemplate)
        print("===================================================================================")
        print("YML FILE MUST BE COMPLETED FOR THE FILE PICKER TO WORK PROPERLY. COMPLETE AND RERUN")
        print("===================================================================================")

        # print(readme)

        sys.exit()

    ################################################################################################
    # FINALIZING FILTERING PARAMETERS ##############################################################
    ################################################################################################

    if re.search(r"^\W+$", ymlDic["miu_specific_uri"]):
        filterParameter = "."
    else:
        filterParameter = ymlDic["miu_specific_uri"]

    filesToOpen = int(ymlDic["files_to_open"])
    pathToKate = ymlDic["path_to_kate"]
    path = ymlDic["miu_main_path"]
    reviewer = ymlDic["reviewer"]
    osVar = ymlDic["system"]

    ################################################################################################
    # LOADING MIU DATA, RANDOMLY SELECTING, OPENING FILES WITH KATE ################################
    ################################################################################################

    dictionary = generateDicOfMIUs(path)
    finalSample = sampleMIUs(dictionary, filterParameter, filesToOpen)

    ################################################################################################
    # OPEN MIUS from THE SELECTED SAMPLE ###########################################################
    ################################################################################################

    counter = filesToOpen

    for f in finalSample:
        with open(dictionary[f], "r", encoding="utf8") as ftIn:
            MIUcontent = ftIn.read()

            if re.search(r"reviewed\s+:\s+(False|NOT REVIEWED)", MIUcontent):
                # insert the name of the reviewer
                if re.search(r"reviewer\s+:", MIUcontent):
                    MIUcontent = re.sub(r"(reviewed\s+:)[^\n]+\n", r"\1 NOT REVIEWED\n", MIUcontent)
                    MIUcontent = re.sub(r"(reviewer\s+:)[^\n]+\n", r"\1 %s\n" % reviewer, MIUcontent)
                else:
                    MIUcontent = re.sub(
                        r"(reviewed)(\s+:)[^\n]+\n", r"\1\2 NOT REVIEWED\nreviewer\2 %s\n" % reviewer, MIUcontent
                        )

                # inserting the name of the reviewer; the reviewer will have to only change the status tag
                # upon saving and closing
                with open(dictionary[f], "w", encoding="utf8") as ftOut:
                    ftOut.write(MIUcontent)

                # NOW OPENING IN KATE FOR REVIEW
                if osVar == "mac":
                    # open on mac
                    lineToRun = "open -a %s %s" % (pathToKate, dictionary[f])
                    os.system(lineToRun)

                elif osVar == "lin":
                    lineToRun = f"{pathToKate} {dictionary[f]}"
                    os.system(lineToRun)
                elif osVar == "win":
                    # open on windows
                    # start path_to_app path_to_file
                    pass
                else:
                    print("Operating system is incorrect. Use: mac, lin, or win")
                    sys.exit()

                print("\t- ", f)

                counter -= 1
                if counter == 0:
                    break

    print("======================================================================================")
    print("REVIEW MIU FILES IN KATE; WHEN DONE, CHANGE STATUS TO `REVIEWED`, SAVE, AND CLOSE ====")
    print("======================================================================================")



from __future__ import annotations

import re

from typing import Any, Dict, Optional, Type

from eis1600.markdown.re_patterns import AR_STR_AND_TAGS, MIU_HEADER
from eis1600.miu.HeadingTracker import HeadingTracker


class YAMLHandler:
    """A class to take care of the MIU YAML Headers

    :param Dict yml: the YAML header as a dict, optional.
    :ivar Literal['NOT REVIEWED', 'REVIEWED'] reviewed: Indicates if the file has manually been reviewed, defaults to
    'NOT REVIEWED'.
    :ivar str reviewer: Initials of the reviewer if the file was already manually reviewed, defaults to None.
    :ivar HeadingTracker headings: HeadingTracker returned by the get_curr_state method of the HeaderTracker.
    :iver List dates: List of dates contained in headings and text
    """

    @staticmethod
    def __parse_yml_val(val: str) -> Any:
        val = val.strip('"\'')
        if val.isdigit():
            return int(val)
        elif val == 'True':
            return True
        elif val == 'False':
            return False
        else:
            return val

    @staticmethod
    def __parse_yml(yml_str: str) -> Dict:
        yml = {}
        level = []
        dict_elem = {}
        # print(yml_str)
        for line in yml_str.splitlines():
            if not line.startswith('#'):
                intend = (len(line) - len(line.lstrip())) / 4
                key_val = line.split(':')
                key = key_val[0].strip(' -')
                val = ':'.join(key_val[1:]).strip()

                if intend < len(level):
                    yml[level[0]] = dict_elem
                    dict_elem = {}
                    level.pop()
                    yml[key] = YAMLHandler.__parse_yml_val(val)
                elif intend and intend == len(level):
                    dict_elem[key] = YAMLHandler.__parse_yml_val(val)
                elif val == '':
                    dict_elem = {}
                    level.append(key)
                else:
                    yml[key] = YAMLHandler.__parse_yml_val(val)

        if len(level):
            yml[level[0]] = dict_elem

        return yml

    def __init__(self, yml: Optional[Dict] = None) -> None:
        self.reviewed = 'NOT REVIEWED'
        self.reviewer = None
        self.headings = None
        self.dates = None

        if yml:
            for key, val in yml.items():
                if key == 'headings':
                    val = HeadingTracker(val)
                self.__setattr__(key, val)

    @classmethod
    def from_yml_str(cls, yml_str: str) -> Type[YAMLHandler]:
        """Return instance with attr set from the yml_str."""
        return cls(YAMLHandler.__parse_yml(yml_str))

    def set_headings(self, headings: Type[HeadingTracker]):
        self.headings = headings

    def get_yamlfied(self) -> str:
        yaml_str = MIU_HEADER + 'Begin#\n\n'
        for key, val in vars(self).items():
            if key == 'dates' and val is not None:
                yaml_str += key + '    : ['
                for date in val:
                    yaml_str += '"' + date + '",'
                yaml_str = yaml_str[:-1]
                yaml_str += ']\n'
            else:
                yaml_str += key + '    : ' + str(val) + '\n'
        yaml_str += '\n' + MIU_HEADER + 'End#\n\n'

        return yaml_str

    def is_reviewed(self) -> bool:
        return self.reviewed == 'REVIEWED'

    def add_date(self, date_tag: str):
        if self.dates:
            self.dates.append(date_tag)
        else:
            self.dates = [date_tag]

    def __setitem__(self, key: str, value: Any) -> None:
        super().__setattr__(key, value)

    def __repr__(self) -> str:
        return str(self.__dict__)

    def __str__(self) -> str:
        return self.get_yamlfied()



import sys
import os
from argparse import ArgumentParser, Action, RawDescriptionHelpFormatter
from multiprocessing import Pool

from eis1600.helper.repo import get_files_from_eis1600_dir, write_to_readme, read_files_from_readme
from eis1600.markdown.methods import update_uids


class CheckFileEndingAction(Action):
    def __call__(self, parser, namespace, input_arg, option_string=None):
        if input_arg and os.path.isfile(input_arg):
            filepath, fileext = os.path.splitext(input_arg)
            if fileext != '.EIS1600':
                parser.error('You need to input an EIS1600 file')
            else:
                setattr(namespace, self.dest, input_arg)
        else:
            setattr(namespace, self.dest, None)


def main():
    arg_parser = ArgumentParser(
            prog=sys.argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to insert UIDs in updated EIS1600 file(s).
-----
Give a single EIS1600 file as input
or 
Use -e <EIS1600_repo> to batch process all EIS1600 files in the EIS1600 directory which have not been processed yet.
'''
    )
    arg_parser.add_argument('-v', '--verbose', action='store_true')
    arg_parser.add_argument(
            '-e', '--eis1600_repo', type=str,
            help='Takes a path to the EIS1600 file repo and batch processes all files which have not been processed yet'
    )
    arg_parser.add_argument(
            'input', type=str, nargs='?',
            help='EIS1600 file to process',
            action=CheckFileEndingAction
    )
    args = arg_parser.parse_args()

    verbose = args.verbose

    if args.input:
        infile = './' + args.input
        if 'data' in infile:
            path = infile.split('data')[0]
        else:
            depth = len(infile.split('/'))
            if depth == 2:
                path = '../../../'
            elif depth == 3:
                path = '../../'
            else:
                path = '../'
        print(f'Update UIDs in {infile}')
        update_uids(infile, verbose)
        infiles = [infile.split('/')[-1]]
        write_to_readme(path, infiles, '# Texts updated with missing UIDs\n')
    elif args.eis1600_repo:
        input_dir = args.eis1600_repo
        if not input_dir[-1] == '/':
            input_dir += '/'

        print(
                f'Insert missing UIDs into checked files from the EIS1600 repo'
        )
        files_list = read_files_from_readme(
                input_dir, '# Texts converted into `.EIS1600`\n', False
        )
        infiles = get_files_from_eis1600_dir(input_dir, files_list, 'EIS1600')
        if not infiles:
            print(
                    'There are no more files to update'
            )
            sys.exit()

        params = [(infile, verbose) for infile in infiles]

        with Pool() as p:
            p.starmap_async(update_uids, params).get()

        write_to_readme(input_dir, infiles, '# Texts updated with missing UIDs\n')
    else:
        print(
                'Pass in a <uri.EIS1600> file to process a single file or use the -e option for batch processing'
        )
        sys.exit()

    print('Done')



from os.path import split, splitext

from eis1600.markdown.UIDs import UIDs
from typing import Optional

import sys
import os
from argparse import ArgumentParser, Action, RawDescriptionHelpFormatter
from multiprocessing import Pool

from eis1600.helper.repo import get_files_from_eis1600_dir, write_to_readme, read_files_from_readme
from eis1600.markdown.re_patterns import BIO_CHR_TO_NEWLINE_PATTERN, HEADER_END_PATTERN, HEADING_OR_BIO_PATTERN, \
    MIU_LIGHT_OR_EIS1600_PATTERN, NEWLINES_CROWD_PATTERN, UID_PATTERN


class CheckFileEndingAction(Action):
    def __call__(self, parser, namespace, input_arg, option_string=None):
        if input_arg and os.path.isfile(input_arg):
            filepath, fileext = os.path.splitext(input_arg)
            if fileext != '.EIS1600':
                parser.error('You need to input an EIS1600 file')
            else:
                setattr(namespace, self.dest, input_arg)
        else:
            setattr(namespace, self.dest, None)


def xx_update_uids(infile: str, verbose: Optional[bool] = False) -> None:
    """Updates a text with missing UIDs.

    :param str infile: Path of the file to update UIDs in.
    :param bool verbose: If True outputs a notification of the file which is currently processed, defaults to False.
    :return None:
    """

    path, ext = splitext(infile)
    outfile = path + '.EIS1600'
    path, uri = split(infile)

    if verbose:
        print(f'Update UIDs in {uri}')

    with open(infile, 'r', encoding='utf8') as infile_h:
        text = infile_h.read()

    text = BIO_CHR_TO_NEWLINE_PATTERN.sub(r'\1\n\2', text)
    header_and_text = HEADER_END_PATTERN.split(text)
    header = header_and_text[0] + header_and_text[1]
    text = header_and_text[2]
    text = NEWLINES_CROWD_PATTERN.sub('\n\n', text)
    text = text.split('\n\n')
    text_updated = []

    used_ids = []

    for paragraph in text:
        if UID_PATTERN.match(paragraph):
            used_ids.append(int(UID_PATTERN.match(paragraph).group('UID')))

    uids = UIDs(used_ids)

    text_iter = text.__iter__()
    paragraph = next(text_iter, None)
    while paragraph is not None:
        next_p = next(text_iter, None)

        if paragraph:
            # Only do this is paragraph is not empty
            if HEADING_OR_BIO_PATTERN.match(paragraph):
                paragraph = paragraph.replace('#', f'_ء_#={uids.get_uid()}=')
                if next_p and not MIU_LIGHT_OR_EIS1600_PATTERN.match(next_p):
                    heading_and_text = paragraph.split('\n', 1)
                    if len(heading_and_text) > 1:
                        paragraph = heading_and_text[0] + f'\n\n_ء_={uids.get_uid()}= ::UNDEFINED:: ~\n' + \
                                    heading_and_text[1]
            elif not UID_PATTERN.match(paragraph):
                section_header = '' if paragraph.startswith('::') else '::UNDEFINED:: ~\n'
                paragraph = f'_ء_={uids.get_uid()}= {section_header}' + paragraph

            text_updated.append(paragraph)

        paragraph = next_p

    text = '\n\n'.join(text_updated)

    # reassemble text
    final = header + '\n\n' + text

    with open(outfile, 'w', encoding='utf8') as outfile_h:
        outfile_h.write(final)


def main():
    arg_parser = ArgumentParser(
            prog=sys.argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to insert UIDs in updated EIS1600 file(s).
-----
Give a single EIS1600 file as input
or 
Use -e <EIS1600_repo> to batch process all EIS1600 files in the EIS1600 directory which have not been processed yet.
'''
    )
    arg_parser.add_argument('-v', '--verbose', action='store_true')
    arg_parser.add_argument(
            '-e', '--eis1600_repo', type=str,
            help='Takes a path to the EIS1600 file repo and batch processes all files which have not been processed yet'
    )
    arg_parser.add_argument(
            'input', type=str, nargs='?',
            help='EIS1600 file to process',
            action=CheckFileEndingAction
    )
    args = arg_parser.parse_args()

    verbose = args.verbose

    if args.input:
        infile = './' + args.input
        xx_update_uids(infile, verbose)
    elif args.eis1600_repo:
        input_dir = args.eis1600_repo
        if not input_dir[-1] == '/':
            input_dir += '/'

        print(
                f'Insert missing UIDs into checked files from the EIS1600 repo (only for EIS1600 files which have not '
                f'been updated yet)'
        )
        files_list = read_files_from_readme(
            input_dir, '# Texts which need to be updated (old process)\n', False
            )
        infiles = get_files_from_eis1600_dir(input_dir, files_list, 'EIS1600')
        if not infiles:
            print(
                    'There are no more files to update'
            )
            sys.exit()

        params = [(infile, verbose) for infile in infiles]

        with Pool() as p:
            p.starmap_async(xx_update_uids, params).get()

        write_to_readme(input_dir, infiles, '# Texts updated with missing UIDs\n')
    else:
        print(
                'Pass in a <uri.EIS1600> file to process a single file or use the -e option for batch processing'
        )
        sys.exit()

    print('Done')



from __future__ import annotations
import copy

from typing import Dict, Optional, Type


class HeadingTracker:
    """A class to keep track of the super elements of a MIU.

    This class keeps track of the headings on different levels to keep this information in the MIU YAML header. Some
    headings are empty stings
    :ivar str h1: Level 1 heading, optional.
    :ivar str h2: Level 2 heading, optional.
    :ivar str h3: Level 3 heading, optional.
    :ivar str h4: Level 4 heading, optional.
    """

    def __init__(self, headings_dict: Optional[Dict] = None) -> None:
        """Constructor which sets attributes to empty strings."""

        self.h1 = None
        self.h2 = None
        self.h3 = None
        self.h4 = None

        if headings_dict:
            for key, val in headings_dict.items():
                self.__setattr__(key, val)

    def get_curr_state(self) -> Type[HeadingTracker]:
        """Get current state of the tacker as deepcopy.

        Returns a deepcopy of the current state.
        :return Type[HeadingTracker]: Deepcopy of the current state of the tracker.
        """

        return copy.deepcopy(self)

    def get_yamlfied(self) -> str:
        """Stringifies HeadingTracker in YAML format, only includes levels which are set.

        :return str: returns the HeadingTracker in YAML format as a string.
        """

        if self.h1 is None:
            return '[]'

        heading_tracker_str = '\n'
        heading_tracker_str += '    - h1    : "' + self.h1 + '"\n'
        if self.h2 is not None:
            heading_tracker_str += '    - h2    : "' + self.h2 + '"\n'
            if self.h3 is not None:
                heading_tracker_str += '    - h3    : "' + self.h3 + '"\n'
                if self.h4 is not None:
                    heading_tracker_str += '    - h4    : "' + self.h4 + '"\n'

        return heading_tracker_str[:-1]     # Return without last newline

    def track(self, level: int, heading: str) -> None:
        """Checks which of the levels changed and sets all sub levels to None (some headings are just an empty string).

        :param int level: The level of the heading indicated by the number of leading `|`.
        :param str heading: The new heading text for the given level.
        """

        if level == 1:
            self.h1 = heading
            self.h2 = None
            self.h3 = None
            self.h4 = None
        elif level == 2:
            self.h2 = heading
            self.h3 = None
            self.h4 = None
        elif level == 3:
            self.h3 = heading
            self.h4 = None
        else:
            self.h4 = heading

    def __repr__(self) -> str:
        return str(self.__dict__)

    def __str__(self) -> str:
        return self.get_yamlfied()
